import inspect
from typing import Any, List

from .dependency_source import Alias, Decorator, DependencySource, Factory
from .exceptions import InvalidGraphError


def is_dependency_source(attribute: Any) -> bool:
    return isinstance(attribute, DependencySource)


class Provider:
    """
    A collection of dependency sources.

    Inherit this class and add attributes using
    `provide`, `alias` or `decorate`.

    You can use `__init__`, regular methods and attributes as usual,
    they won't be analyzed when creating a container

    The only intended usage of providers is to pass them when
    creating a container
    """

    def __init__(self):
        self.factories: List[Factory] = []
        self.aliases: List[Alias] = []
        self.decorators: List[Decorator] = []
        self._init_dependency_sources()

    def _init_dependency_sources(self) -> None:
        processed_types = {}

        source: DependencySource
        for name, source in inspect.getmembers(self, is_dependency_source):
            if source.provides in processed_types:
                raise InvalidGraphError(
                    f"Type {source.provides} is registered multiple times "
                    f"in the same {Provider} by attributes "
                    f"{processed_types[source.provides]!r} and {name!r}",
                )
            if isinstance(source, Alias):
                self.aliases.append(source)
            if isinstance(source, Factory):
                self.factories.append(source)
            if isinstance(source, Decorator):
                self.decorators.append(source)
            processed_types[source.provides] = name



__all__ = [
    "make_async_container", "AsyncContainer",
    "make_container", "Container",
    "Depends", "wrap_injection",
    "Provider",
    "alias", "decorate", "provide",
    "BaseScope", "Scope",
]

from .async_container import AsyncContainer, make_async_container
from .container import Container, make_container
from .dependency_source import alias, decorate, provide
from .inject import Depends, wrap_injection
from .provider import Provider
from .scope import BaseScope, Scope



try:
    from builtins import ExceptionGroup
except ImportError:
    from exceptiongroup import ExceptionGroup


class DishkaException:
    pass


class InvalidGraphError(DishkaException, ValueError):
    pass


class ExitExceptionGroup(ExceptionGroup, DishkaException):
    pass



from dataclasses import dataclass
from threading import Lock
from typing import Callable, List, Optional, Type, TypeVar

from .dependency_source import Factory, FactoryType
from .exceptions import ExitExceptionGroup
from .provider import Provider
from .registry import Registry, make_registries
from .scope import BaseScope, Scope

T = TypeVar("T")


@dataclass
class Exit:
    __slots__ = ("type", "callable")
    type: FactoryType
    callable: Callable


class Container:
    __slots__ = (
        "registry", "child_registries", "context", "parent_container",
        "lock", "_exits",
    )

    def __init__(
            self,
            registry: Registry,
            *child_registries: Registry,
            parent_container: Optional["Container"] = None,
            context: Optional[dict] = None,
            lock_factory: Callable[[], Lock] | None = None,
    ):
        self.registry = registry
        self.child_registries = child_registries
        self.context = {type(self): self}
        if context:
            self.context.update(context)
        self.parent_container = parent_container
        if lock_factory:
            self.lock = lock_factory()
        else:
            self.lock = None
        self._exits: List[Exit] = []

    def _create_child(
            self,
            context: Optional[dict],
            lock_factory: Callable[[], Lock] | None,
    ) -> "Container":
        return Container(
            *self.child_registries,
            parent_container=self,
            context=context,
            lock_factory=lock_factory,
        )

    def __call__(
            self,
            context: Optional[dict] = None,
            lock_factory: Callable[[], Lock] | None = None,
    ) -> "ContextWrapper":
        """
        Prepare container for entering the inner scope.
        :param context: Data which will available in inner scope
        :param lock_factory: Callable to create lock instance or None
        :return: context manager for inner scope
        """
        if not self.child_registries:
            raise ValueError("No child scopes found")
        return ContextWrapper(self._create_child(context, lock_factory))

    def _get_from_self(self, factory: Factory) -> T:
        sub_dependencies = [
            self._get_unlocked(dependency)
            for dependency in factory.dependencies
        ]
        if factory.type is FactoryType.GENERATOR:
            generator = factory.source(*sub_dependencies)
            self._exits.append(Exit(factory.type, generator))
            return next(generator)
        elif factory.type is FactoryType.FACTORY:
            return factory.source(*sub_dependencies)
        elif factory.type is FactoryType.VALUE:
            return factory.source
        else:
            raise ValueError(f"Unsupported type {factory.type}")

    def get(self, dependency_type: Type[T]) -> T:
        lock = self.lock
        if not lock:
            return self._get_unlocked(dependency_type)
        with lock:
            return self._get_unlocked(dependency_type)

    def _get_unlocked(self, dependency_type: Type[T]) -> T:
        if dependency_type in self.context:
            return self.context[dependency_type]
        provider = self.registry.get_provider(dependency_type)
        if not provider:
            if not self.parent_container:
                raise ValueError(f"No provider found for {dependency_type!r}")
            return self.parent_container.get(dependency_type)
        solved = self._get_from_self(provider)
        self.context[dependency_type] = solved
        return solved

    def close(self) -> None:
        errors = []
        for exit_generator in self._exits[::-1]:
            try:
                if exit_generator.type is FactoryType.GENERATOR:
                    next(exit_generator.callable)
            except StopIteration:
                pass
            except Exception as err:  # noqa: BLE001
                errors.append(err)
        if errors:
            raise ExitExceptionGroup("Cleanup context errors", errors)


class ContextWrapper:
    __slots__ = ("container",)

    def __init__(self, container: Container):
        self.container = container

    def __enter__(self) -> Container:
        return self.container

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.container.close()


def make_container(
        *providers: Provider,
        scopes: Type[BaseScope] = Scope,
        context: Optional[dict] = None,
        lock_factory: Callable[[], Lock] | None = None,
) -> ContextWrapper:
    registries = make_registries(*providers, scopes=scopes)
    return ContextWrapper(
        Container(*registries, context=context, lock_factory=lock_factory),
    )



from collections import defaultdict
from typing import Any, List, NewType, Type

from .dependency_source import Factory
from .exceptions import InvalidGraphError
from .provider import Provider
from .scope import BaseScope


class Registry:
    __slots__ = ("scope", "_factories")

    def __init__(self, scope: BaseScope):
        self._factories: dict[Type, Factory] = {}
        self.scope = scope

    def add_provider(self, factory: Factory):
        self._factories[factory.provides] = factory

    def get_provider(self, dependency: Any) -> Factory:
        return self._factories.get(dependency)


def make_registries(
        *providers: Provider, scopes: Type[BaseScope],
) -> List[Registry]:
    dep_scopes: dict[Type, BaseScope] = {}
    alias_sources = {}
    for provider in providers:
        for source in provider.factories:
            dep_scopes[source.provides] = source.scope
        for source in provider.aliases:
            alias_sources[source.provides] = source.source

    registries = {scope: Registry(scope) for scope in scopes}
    decorator_depth: dict[Type, int] = defaultdict(int)

    for provider in providers:
        for source in provider.factories:
            scope = source.scope
            registries[scope].add_provider(source)
        for source in provider.aliases:
            alias_source = source.source
            visited_types = [alias_source]
            while alias_source not in dep_scopes:
                alias_source = alias_sources[alias_source]
                if alias_source in visited_types:
                    raise InvalidGraphError(
                        f"Cycle aliases detected {visited_types}",
                    )
                visited_types.append(alias_source)
            scope = dep_scopes[alias_source]
            dep_scopes[source.provides] = scope
            source = source.as_factory(scope)
            registries[scope].add_provider(source)
        for source in provider.decorators:
            provides = source.provides
            scope = dep_scopes[provides]
            registry = registries[scope]
            undecorated_type = NewType(
                f"{provides.__name__}@{decorator_depth[provides]}",
                source.provides,
            )
            decorator_depth[provides] += 1
            old_provider = registry.get_provider(provides)
            old_provider.provides = undecorated_type
            registry.add_provider(old_provider)
            source = source.as_factory(
                scope, undecorated_type,
            )
            registries[scope].add_provider(source)

    return list(registries.values())



import pytest

from dishka import Provider, Scope, alias, provide
from dishka.dependency_source import FactoryType
from .sample_providers import (
    ClassA,
    async_func_a,
    async_gen_a,
    async_iter_a,
    sync_func_a,
    sync_gen_a,
    sync_iter_a,
)


def test_provider_init():
    class MyProvider(Provider):
        a = alias(source=int, provides=bool)

        @provide(scope=Scope.REQUEST)
        def foo(self, x: bool) -> str:
            return f"{x}"

    provider = MyProvider()
    assert len(provider.factories) == 1
    assert len(provider.aliases) == 1


@pytest.mark.parametrize(
    "source, provider_type, is_to_bound", [
        (ClassA, FactoryType.FACTORY, False),
        (sync_func_a, FactoryType.FACTORY, True),
        (sync_iter_a, FactoryType.GENERATOR, True),
        (sync_gen_a, FactoryType.GENERATOR, True),
        (async_func_a, FactoryType.ASYNC_FACTORY, True),
        (async_iter_a, FactoryType.ASYNC_GENERATOR, True),
        (async_gen_a, FactoryType.ASYNC_GENERATOR, True),
    ],
)
def test_parse_factory(source, provider_type, is_to_bound):
    factory = provide(source, scope=Scope.REQUEST)
    assert factory.provides == ClassA
    assert factory.dependencies == [int]
    assert factory.is_to_bound == is_to_bound
    assert factory.scope == Scope.REQUEST
    assert factory.source == source
    assert factory.type == provider_type



from inspect import Parameter, Signature, signature
from typing import (
    Annotated,
    Any,
    Callable,
    Sequence,
    get_args,
    get_origin,
    get_type_hints,
)

from .container import Container


class Depends:
    def __init__(self, param: Any = None):
        self.param = param


def default_parse_dependency(
        parameter: Parameter,
        hint: Any,
) -> Any:
    """ Resolve dependency type or return None if it is not a dependency """
    if get_origin(hint) is not Annotated:
        return None
    dep = next(
        (arg for arg in get_args(hint) if isinstance(arg, Depends)),
        None,
    )
    if not dep:
        return None
    if dep.param is None:
        return get_args(hint)[0]
    else:
        return dep.param


DependencyParser = Callable[[Parameter, Any], Any]


def wrap_injection(
        func: Callable,
        container_getter: Callable[[dict], Container],
        remove_depends: bool = True,
        additional_params: Sequence[Parameter] = (),
        is_async: bool = False,
        parse_dependency: DependencyParser = default_parse_dependency,
) -> Callable:
    hints = get_type_hints(func, include_extras=True)
    func_signature = signature(func)

    dependencies = {}
    for name, param in func_signature.parameters.items():
        hint = hints.get(name, Any)
        dep = parse_dependency(param, hint)
        if dep is None:
            continue
        dependencies[name] = dep

    if remove_depends:
        new_annotations = {
            name: hint
            for name, hint in hints.items()
            if name not in dependencies
        }
        new_params = [
            param
            for name, param in func_signature.parameters.items()
            if name not in dependencies
        ]
    else:
        new_annotations = hints.copy()
        new_params = func_signature.parameters.copy()

    if additional_params:
        new_params.extend(additional_params)
        for param in additional_params:
            new_annotations[param.name] = param.annotation

    if is_async:
        async def autoinjected_func(*args, **kwargs):
            container = container_getter(kwargs)
            for param in additional_params:
                kwargs.pop(param.name)
            solved = {
                name: await container.get(dep)
                for name, dep in dependencies.items()
            }
            return await func(*args, **kwargs, **solved)
    else:
        def autoinjected_func(*args, **kwargs):
            container = container_getter(kwargs)
            for param in additional_params:
                kwargs.pop(param.name)
            solved = {
                name: container.get(dep)
                for name, dep in dependencies.items()
            }
            return func(*args, **kwargs, **solved)

    autoinjected_func.__name__ = func.__name__
    autoinjected_func.__doc__ = func.__doc__
    autoinjected_func.__annotations__ = new_annotations
    autoinjected_func.__signature__ = Signature(
        parameters=new_params,
        return_annotation=func_signature.return_annotation,
    )
    return autoinjected_func



from collections.abc import AsyncIterable, Iterable
from enum import Enum
from inspect import (
    isasyncgenfunction,
    isclass,
    iscoroutinefunction,
    isgeneratorfunction,
)
from typing import (
    Any,
    Callable,
    Optional,
    Sequence,
    Type,
    get_args,
    get_origin,
    get_type_hints,
    overload,
)

from .scope import BaseScope


class FactoryType(Enum):
    GENERATOR = "generator"
    ASYNC_GENERATOR = "async_generator"
    FACTORY = "factory"
    ASYNC_FACTORY = "async_factory"
    VALUE = "value"


def _identity(x: Any) -> Any:
    return x


class Factory:
    __slots__ = (
        "dependencies", "source", "provides", "scope", "type",
        "is_to_bound",
    )

    def __init__(
            self,
            dependencies: Sequence[Any],
            source: Any,
            provides: Type,
            scope: Optional[BaseScope],
            type: FactoryType,
            is_to_bound: bool,
    ):
        self.dependencies = dependencies
        self.source = source
        self.provides = provides
        self.scope = scope
        self.type = type
        self.is_to_bound = is_to_bound

    def __get__(self, instance, owner):
        if instance is None:
            return self
        if self.is_to_bound:
            source = self.source.__get__(instance, owner)
        else:
            source = self.source
        return Factory(
            dependencies=self.dependencies,
            source=source,
            provides=self.provides,
            scope=self.scope,
            type=self.type,
            is_to_bound=False,
        )


def make_factory(
        provides: Any,
        scope: Optional[BaseScope],
        source: Callable,
) -> Factory:
    if isclass(source):
        hints = get_type_hints(source.__init__, include_extras=True)
        hints.pop("return", None)
        possible_dependency = source
        is_to_bind = False
    else:
        hints = get_type_hints(source, include_extras=True)
        possible_dependency = hints.pop("return", None)
        is_to_bind = True

    if isclass(source):
        provider_type = FactoryType.FACTORY
    elif isasyncgenfunction(source):
        provider_type = FactoryType.ASYNC_GENERATOR
        if get_origin(possible_dependency) is AsyncIterable:
            possible_dependency = get_args(possible_dependency)[0]
        else:  # async generator
            possible_dependency = get_args(possible_dependency)[0]
    elif isgeneratorfunction(source):
        provider_type = FactoryType.GENERATOR
        if get_origin(possible_dependency) is Iterable:
            possible_dependency = get_args(possible_dependency)[0]
        else:  # generator
            possible_dependency = get_args(possible_dependency)[1]
    elif iscoroutinefunction(source):
        provider_type = FactoryType.ASYNC_FACTORY
    else:
        provider_type = FactoryType.FACTORY

    return Factory(
        dependencies=list(hints.values()),
        type=provider_type,
        source=source,
        scope=scope,
        provides=provides or possible_dependency,
        is_to_bound=is_to_bind,
    )


@overload
def provide(
        *,
        scope: BaseScope,
        provides: Any = None,
) -> Callable[[Callable], Factory]:
    ...


@overload
def provide(
        source: Callable | Type,
        *,
        scope: BaseScope,
        provides: Any = None,
) -> Factory:
    ...


def provide(
        source: Callable | Type | None = None,
        *,
        scope: BaseScope,
        provides: Any = None,
) -> Factory | Callable[[Callable], Factory]:
    """
    Mark a method or class as providing some dependency.

    If used as a method decorator then return annotation is used
    to determine what is provided. User `provides` to override that.
    Method parameters are analyzed and passed automatically.

    If used with a class a first parameter than `__init__` method parameters
    are passed automatically. If no provides is passed then it is
    supposed that class itself is a provided dependency.

    Return value must be saved as a `Provider` class attribute and
    not intended for direct usage

    :param source: Method to decorate or class.
    :param scope: Scope of the dependency to limit its lifetime
    :param provides: Dependency type which is provided by this factory
    :return: instance of Factory or a decorator returning it
    """
    if source is not None:
        return make_factory(provides, scope, source)

    def scoped(func):
        return make_factory(provides, scope, func)

    return scoped


class Alias:
    __slots__ = ("source", "provides")

    def __init__(self, source, provides):
        self.source = source
        self.provides = provides

    def as_factory(self, scope: BaseScope) -> Factory:
        return Factory(
            scope=scope,
            source=_identity,
            provides=self.provides,
            is_to_bound=False,
            dependencies=[self.source],
            type=FactoryType.FACTORY,
        )

    def __get__(self, instance, owner):
        return self


def alias(
        *,
        source: Type,
        provides: Type,
) -> Alias:
    return Alias(
        source=source,
        provides=provides,
    )


class Decorator:
    __slots__ = ("provides", "factory")

    def __init__(self, factory: Factory):
        self.factory = factory
        self.provides = factory.provides

    def as_factory(
            self, scope: BaseScope, new_dependency: Any,
    ) -> Factory:
        return Factory(
            scope=scope,
            source=self.factory.source,
            provides=self.factory.provides,
            is_to_bound=self.factory.is_to_bound,
            dependencies=[
                new_dependency if dep is self.provides else dep
                for dep in self.factory.dependencies
            ],
            type=self.factory.type,
        )

    def __get__(self, instance, owner):
        return Decorator(self.factory.__get__(instance, owner))


@overload
def decorate(
        *,
        provides: Any = None,
) -> Callable[[Callable], Decorator]:
    ...


@overload
def decorate(
        source: Callable | Type,
        *,
        provides: Any = None,
) -> Decorator:
    ...


def decorate(
        source: Callable | Type | None = None,
        provides: Any = None,
) -> Decorator | Callable[[Callable], Decorator]:
    if source is not None:
        return Decorator(make_factory(provides, None, source))

    def scoped(func):
        return Decorator(make_factory(provides, None, func))

    return scoped


DependencySource = Alias | Factory | Decorator
